@@ -31,7 +31,6 @@ import (
 	"sync"
 	"sync/atomic"
 	"time"
-	"unsafe"
 
 	"github.com/pion/sdp/v3"
 )
@@ -46,10 +45,10 @@ var (
 	PathToOcsSignalingBackend = "ocs/v2.php/apps/spreed/api/v1/signaling/backend"
 )
 
-type ClientSession struct {
-	roomJoinTime int64
-	inCall       uint32
+// ResponseHandlerFunc will return "true" has been fully processed.
+type ResponseHandlerFunc func(message *ClientMessage) bool
 
+type ClientSession struct {
 	hub       *Hub
 	events    AsyncEvents
 	privateId string
@@ -61,6 +60,7 @@ type ClientSession struct {
 	userId     string
 	userData   *json.RawMessage
 
+	inCall              Flags
 	supportsPermissions bool
 	permissions         map[Permission]bool
 
@@ -73,7 +73,8 @@ type ClientSession struct {
 	mu sync.Mutex
 
 	client        *Client
-	room          unsafe.Pointer
+	room          atomic.Pointer[Room]
+	roomJoinTime  atomic.Int64
 	roomSessionId string
 
 	publisherWaiters ChannelWaiters
@@ -89,6 +90,9 @@ type ClientSession struct {
 
 	seenJoinedLock   sync.Mutex
 	seenJoinedEvents map[string]bool
+
+	responseHandlersLock sync.Mutex
+	responseHandlers     map[string]ResponseHandlerFunc
 }
 
 func NewClientSession(hub *Hub, privateId string, publicId string, data *SessionIdData, backend *Backend, hello *HelloClientMessage, auth *BackendClientAuthResponse) (*ClientSession, error) {
@@ -165,7 +169,7 @@ func (s *ClientSession) ClientType() string {
 
 // GetInCall is only used for internal clients.
 func (s *ClientSession) GetInCall() int {
-	return int(atomic.LoadUint32(&s.inCall))
+	return int(s.inCall.Get())
 }
 
 func (s *ClientSession) SetInCall(inCall int) bool {
@@ -173,16 +177,7 @@ func (s *ClientSession) SetInCall(inCall int) bool {
 		inCall = 0
 	}
 
-	for {
-		old := atomic.LoadUint32(&s.inCall)
-		if old == uint32(inCall) {
-			return false
-		}
-
-		if atomic.CompareAndSwapUint32(&s.inCall, old, uint32(inCall)) {
-			return true
-		}
-	}
+	return s.inCall.Set(uint32(inCall))
 }
 
 func (s *ClientSession) GetFeatures() []string {
@@ -334,11 +329,11 @@ func (s *ClientSession) IsExpired(now time.Time) bool {
 }
 
 func (s *ClientSession) SetRoom(room *Room) {
-	atomic.StorePointer(&s.room, unsafe.Pointer(room))
+	s.room.Store(room)
 	if room != nil {
-		atomic.StoreInt64(&s.roomJoinTime, time.Now().UnixNano())
+		s.roomJoinTime.Store(time.Now().UnixNano())
 	} else {
-		atomic.StoreInt64(&s.roomJoinTime, 0)
+		s.roomJoinTime.Store(0)
 	}
 
 	s.seenJoinedLock.Lock()
@@ -347,11 +342,11 @@ func (s *ClientSession) SetRoom(room *Room) {
 }
 
 func (s *ClientSession) GetRoom() *Room {
-	return (*Room)(atomic.LoadPointer(&s.room))
+	return s.room.Load()
 }
 
 func (s *ClientSession) getRoomJoinTime() time.Time {
-	t := atomic.LoadInt64(&s.roomJoinTime)
+	t := s.roomJoinTime.Load()
 	if t == 0 {
 		return time.Time{}
 	}
@@ -417,6 +412,36 @@ func (s *ClientSession) SubscribeEvents() error {
 	return s.events.RegisterSessionListener(s.publicId, s.backend, s)
 }
 
+func (s *ClientSession) UpdateRoomSessionId(roomSessionId string) error {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	if s.roomSessionId == roomSessionId {
+		return nil
+	}
+
+	if err := s.hub.roomSessions.SetRoomSession(s, roomSessionId); err != nil {
+		return err
+	}
+
+	if roomSessionId != "" {
+		if room := s.GetRoom(); room != nil {
+			log.Printf("Session %s updated room session id to %s in room %s", s.PublicId(), roomSessionId, room.Id())
+		} else {
+			log.Printf("Session %s updated room session id to %s in unknown room", s.PublicId(), roomSessionId)
+		}
+	} else {
+		if room := s.GetRoom(); room != nil {
+			log.Printf("Session %s cleared room session id in room %s", s.PublicId(), room.Id())
+		} else {
+			log.Printf("Session %s cleared room session id in unknown room", s.PublicId())
+		}
+	}
+
+	s.roomSessionId = roomSessionId
+	return nil
+}
+
 func (s *ClientSession) SubscribeRoomEvents(roomid string, roomSessionId string) error {
 	s.mu.Lock()
 	defer s.mu.Unlock()
@@ -1390,3 +1415,38 @@ func (s *ClientSession) GetVirtualSessions() []*VirtualSession {
 	}
 	return result
 }
+
+func (s *ClientSession) HandleResponse(id string, handler ResponseHandlerFunc) {
+	s.responseHandlersLock.Lock()
+	defer s.responseHandlersLock.Unlock()
+
+	if s.responseHandlers == nil {
+		s.responseHandlers = make(map[string]ResponseHandlerFunc)
+	}
+
+	s.responseHandlers[id] = handler
+}
+
+func (s *ClientSession) ClearResponseHandler(id string) {
+	s.responseHandlersLock.Lock()
+	defer s.responseHandlersLock.Unlock()
+
+	delete(s.responseHandlers, id)
+}
+
+func (s *ClientSession) ProcessResponse(message *ClientMessage) bool {
+	id := message.Id
+	if id == "" {
+		return false
+	}
+
+	s.responseHandlersLock.Lock()
+	cb, found := s.responseHandlers[id]
+	defer s.responseHandlersLock.Unlock()
+
+	if !found {
+		return false
+	}
+
+	return cb(message)
+}
