@@ -97,6 +97,9 @@ var (
 
 	// Delay after which a screen publisher should be cleaned up.
 	cleanupScreenPublisherDelay = time.Second
+
+	// Delay after which a "cleared" / "rejected" dialout status should be removed.
+	removeCallStatusTTL = 5 * time.Second
 )
 
 const (
@@ -109,9 +112,6 @@ func init() {
 }
 
 type Hub struct {
-	// 64-bit members that are accessed atomically must be 64-bit aligned.
-	sid uint64
-
 	events       AsyncEvents
 	upgrader     websocket.Upgrader
 	cookie       *securecookie.SecureCookie
@@ -120,8 +120,8 @@ type Hub struct {
 	welcome      atomic.Value // *ServerMessage
 
 	closer          *Closer
-	readPumpActive  uint32
-	writePumpActive uint32
+	readPumpActive  atomic.Int32
+	writePumpActive atomic.Int32
 
 	roomUpdated      chan *BackendServerRoomRequest
 	roomDeleted      chan *BackendServerRoomRequest
@@ -131,6 +131,7 @@ type Hub struct {
 	mu sync.RWMutex
 	ru sync.RWMutex
 
+	sid      atomic.Uint64
 	clients  map[uint64]*Client
 	sessions map[uint64]Session
 	rooms    map[string]*Room
@@ -150,13 +151,14 @@ type Hub struct {
 	expiredSessions    map[Session]bool
 	anonymousSessions  map[*ClientSession]time.Time
 	expectHelloClients map[*Client]time.Time
+	dialoutSessions    map[*ClientSession]bool
 
 	backendTimeout time.Duration
 	backend        *BackendClient
 
 	geoip          *GeoLookup
 	geoipOverrides map[*net.IPNet]string
-	geoipUpdating  int32
+	geoipUpdating  atomic.Bool
 
 	rpcServer  *GrpcServer
 	rpcClients *GrpcClients
@@ -256,53 +258,52 @@ func NewHub(config *goconf.ConfigFile, events AsyncEvents, rpcServer *GrpcServer
 		if err != nil {
 			return nil, err
 		}
-
-		if options, _ := config.GetOptions("geoip-overrides"); len(options) > 0 {
-			geoipOverrides = make(map[*net.IPNet]string)
-			for _, option := range options {
-				var ip net.IP
-				var ipNet *net.IPNet
-				if strings.Contains(option, "/") {
-					_, ipNet, err = net.ParseCIDR(option)
-					if err != nil {
-						return nil, fmt.Errorf("could not parse CIDR %s: %s", option, err)
-					}
-				} else {
-					ip = net.ParseIP(option)
-					if ip == nil {
-						return nil, fmt.Errorf("could not parse IP %s", option)
-					}
-
-					var mask net.IPMask
-					if ipv4 := ip.To4(); ipv4 != nil {
-						mask = net.CIDRMask(32, 32)
-					} else {
-						mask = net.CIDRMask(128, 128)
-					}
-					ipNet = &net.IPNet{
-						IP:   ip,
-						Mask: mask,
-					}
-				}
-
-				value, _ := config.GetString("geoip-overrides", option)
-				value = strings.ToUpper(strings.TrimSpace(value))
-				if value == "" {
-					log.Printf("IP %s doesn't have a country assigned, skipping", option)
-					continue
-				} else if !IsValidCountry(value) {
-					log.Printf("Country %s for IP %s is invalid, skipping", value, option)
-					continue
-				}
-
-				log.Printf("Using country %s for %s", value, ipNet)
-				geoipOverrides[ipNet] = value
-			}
-		}
 	} else {
 		log.Printf("Not using GeoIP database")
 	}
 
+	if options, _ := GetStringOptions(config, "geoip-overrides", true); len(options) > 0 {
+		geoipOverrides = make(map[*net.IPNet]string, len(options))
+		for option, value := range options {
+			var ip net.IP
+			var ipNet *net.IPNet
+			if strings.Contains(option, "/") {
+				_, ipNet, err = net.ParseCIDR(option)
+				if err != nil {
+					return nil, fmt.Errorf("could not parse CIDR %s: %s", option, err)
+				}
+			} else {
+				ip = net.ParseIP(option)
+				if ip == nil {
+					return nil, fmt.Errorf("could not parse IP %s", option)
+				}
+
+				var mask net.IPMask
+				if ipv4 := ip.To4(); ipv4 != nil {
+					mask = net.CIDRMask(32, 32)
+				} else {
+					mask = net.CIDRMask(128, 128)
+				}
+				ipNet = &net.IPNet{
+					IP:   ip,
+					Mask: mask,
+				}
+			}
+
+			value = strings.ToUpper(strings.TrimSpace(value))
+			if value == "" {
+				log.Printf("IP %s doesn't have a country assigned, skipping", option)
+				continue
+			} else if !IsValidCountry(value) {
+				log.Printf("Country %s for IP %s is invalid, skipping", value, option)
+				continue
+			}
+
+			log.Printf("Using country %s for %s", value, ipNet)
+			geoipOverrides[ipNet] = value
+		}
+	}
+
 	hub := &Hub{
 		events: events,
 		upgrader: websocket.Upgrader{
@@ -338,6 +339,7 @@ func NewHub(config *goconf.ConfigFile, events AsyncEvents, rpcServer *GrpcServer
 		expiredSessions:    make(map[Session]bool),
 		anonymousSessions:  make(map[*ClientSession]time.Time),
 		expectHelloClients: make(map[*Client]time.Time),
+		dialoutSessions:    make(map[*ClientSession]bool),
 
 		backendTimeout: backendTimeout,
 		backend:        backend,
@@ -409,12 +411,12 @@ func (h *Hub) updateGeoDatabase() {
 		return
 	}
 
-	if !atomic.CompareAndSwapInt32(&h.geoipUpdating, 0, 1) {
+	if !h.geoipUpdating.CompareAndSwap(false, true) {
 		// Already updating
 		return
 	}
 
-	defer atomic.CompareAndSwapInt32(&h.geoipUpdating, 1, 0)
+	defer h.geoipUpdating.Store(false)
 	delay := time.Second
 	for !h.closer.IsClosed() {
 		err := h.geoip.Update()
@@ -578,6 +580,24 @@ func (h *Hub) GetSessionByPublicId(sessionId string) Session {
 	return session
 }
 
+func (h *Hub) GetDialoutSession(roomId string, backend *Backend) *ClientSession {
+	url := backend.Url()
+
+	h.mu.RLock()
+	defer h.mu.RUnlock()
+	for session := range h.dialoutSessions {
+		if session.backend.Url() != url {
+			continue
+		}
+
+		if session.GetClient() != nil {
+			return session
+		}
+	}
+
+	return nil
+}
+
 func (h *Hub) checkExpiredSessions(now time.Time) {
 	for s := range h.expiredSessions {
 		if s.IsExpired(now) {
@@ -641,6 +661,7 @@ func (h *Hub) removeSession(session Session) (removed bool) {
 	delete(h.expiredSessions, session)
 	if session, ok := session.(*ClientSession); ok {
 		delete(h.anonymousSessions, session)
+		delete(h.dialoutSessions, session)
 	}
 	h.mu.Unlock()
 	return
@@ -693,9 +714,9 @@ func (h *Hub) sendWelcome(client *Client) {
 }
 
 func (h *Hub) newSessionIdData(backend *Backend) *SessionIdData {
-	sid := atomic.AddUint64(&h.sid, 1)
+	sid := h.sid.Add(1)
 	for sid == 0 {
-		sid = atomic.AddUint64(&h.sid, 1)
+		sid = h.sid.Add(1)
 	}
 	sessionIdData := &SessionIdData{
 		Sid:       sid,
@@ -719,10 +740,6 @@ func (h *Hub) processRegister(client *Client, message *ClientMessage, backend *B
 		return
 	}
 
-	sid := atomic.AddUint64(&h.sid, 1)
-	for sid == 0 {
-		sid = atomic.AddUint64(&h.sid, 1)
-	}
 	sessionIdData := h.newSessionIdData(backend)
 	privateSessionId, err := h.encodeSessionId(sessionIdData, privateSessionName)
 	if err != nil {
@@ -758,7 +775,8 @@ func (h *Hub) processRegister(client *Client, message *ClientMessage, backend *B
 	}
 
 	if limit := uint32(backend.Limit()); limit > 0 && h.rpcClients != nil {
-		totalCount := uint32(backend.Len())
+		var totalCount atomic.Uint32
+		totalCount.Add(uint32(backend.Len()))
 		var wg sync.WaitGroup
 		ctx, cancel := context.WithTimeout(context.Background(), time.Second)
 		defer cancel()
@@ -775,12 +793,12 @@ func (h *Hub) processRegister(client *Client, message *ClientMessage, backend *B
 
 				if count > 0 {
 					log.Printf("%d sessions connected for %s on %s", count, backend.Url(), c.Target())
-					atomic.AddUint32(&totalCount, count)
+					totalCount.Add(count)
 				}
 			}(client)
 		}
 		wg.Wait()
-		if totalCount > limit {
+		if totalCount.Load() > limit {
 			backend.RemoveSession(session)
 			log.Printf("Error adding session %s to backend %s: %s", session.PublicId(), backend.Id(), SessionLimitExceeded)
 			session.Close()
@@ -802,8 +820,12 @@ func (h *Hub) processRegister(client *Client, message *ClientMessage, backend *B
 	h.sessions[sessionIdData.Sid] = session
 	h.clients[sessionIdData.Sid] = client
 	delete(h.expectHelloClients, client)
-	if userId == "" && auth.Type != HelloClientTypeInternal {
+	if userId == "" && session.ClientType() != HelloClientTypeInternal {
 		h.startWaitAnonymousSessionRoomLocked(session)
+	} else if session.ClientType() == HelloClientTypeInternal && session.HasFeature(ClientFeatureStartDialout) {
+		// TODO: There is a small race condition for sessions that take some time
+		// between connecting and joining a room.
+		h.dialoutSessions[session] = true
 	}
 	h.mu.Unlock()
 
@@ -1250,12 +1272,31 @@ func (h *Hub) processRoom(client *Client, message *ClientMessage) {
 				h.startWaitAnonymousSessionRoom(session)
 			}
 		}
+
 		return
 	}
 
 	if session != nil {
 		if room := h.getRoomForBackend(roomId, session.Backend()); room != nil && room.HasSession(session) {
 			// Session already is in that room, no action needed.
+			roomSessionId := message.Room.SessionId
+			if roomSessionId == "" {
+				// TODO(jojo): Better make the session id required in the request.
+				log.Printf("User did not send a room session id, assuming session %s", session.PublicId())
+				roomSessionId = session.PublicId()
+			}
+
+			if err := session.UpdateRoomSessionId(roomSessionId); err != nil {
+				log.Printf("Error updating room session id for session %s: %s", session.PublicId(), err)
+			}
+			session.SendMessage(message.NewErrorServerMessage(
+				NewErrorDetail("already_joined", "Already joined this room.", &RoomErrorDetails{
+					Room: &RoomServerMessage{
+						RoomId:     room.id,
+						Properties: room.properties,
+					},
+				}),
+			))
 			return
 		}
 	}
@@ -1371,6 +1412,10 @@ func (h *Hub) processJoinRoom(session *ClientSession, message *ClientMessage, ro
 	h.mu.Lock()
 	// The session now joined a room, don't expire if it is anonymous.
 	delete(h.anonymousSessions, session)
+	if session.ClientType() == HelloClientTypeInternal && session.HasFeature(ClientFeatureStartDialout) {
+		// An internal session in a room can not be used for dialout.
+		delete(h.dialoutSessions, session)
+	}
 	h.mu.Unlock()
 	session.SetRoom(r)
 	if room.Room.Permissions != nil {
@@ -1754,6 +1799,10 @@ func (h *Hub) processInternalMsg(client *Client, message *ClientMessage) {
 		return
 	}
 
+	if session.ProcessResponse(message) {
+		return
+	}
+
 	switch msg.Type {
 	case "addsession":
 		msg := msg.AddSession
@@ -1905,6 +1954,38 @@ func (h *Hub) processInternalMsg(client *Client, message *ClientMessage) {
 				room.NotifySessionChanged(session, SessionChangeInCall)
 			}
 		}
+	case "dialout":
+		roomId := msg.Dialout.RoomId
+		msg.Dialout.RoomId = "" // Don't send room id to recipients.
+		if msg.Dialout.Type == "status" {
+			asyncMessage := &AsyncMessage{
+				Type: "room",
+				Room: &BackendServerRoomRequest{
+					Type: "transient",
+					Transient: &BackendRoomTransientRequest{
+						Action: TransientActionSet,
+						Key:    "callstatus_" + msg.Dialout.Status.CallId,
+						Value:  msg.Dialout.Status,
+					},
+				},
+			}
+			if msg.Dialout.Status.Status == DialoutStatusCleared || msg.Dialout.Status.Status == DialoutStatusRejected {
+				asyncMessage.Room.Transient.TTL = removeCallStatusTTL
+			}
+			if err := h.events.PublishBackendRoomMessage(roomId, session.Backend(), asyncMessage); err != nil {
+				log.Printf("Error publishing dialout message %+v to room %s", msg.Dialout, roomId)
+			}
+		} else {
+			if err := h.events.PublishRoomMessage(roomId, session.Backend(), &AsyncMessage{
+				Type: "message",
+				Message: &ServerMessage{
+					Type:    "dialout",
+					Dialout: msg.Dialout,
+				},
+			}); err != nil {
+				log.Printf("Error publishing dialout message %+v to room %s", msg.Dialout, roomId)
+			}
+		}
 	default:
 		log.Printf("Ignore unsupported internal message %+v from %s", msg, session.PublicId())
 		return
@@ -1947,9 +2028,9 @@ func (h *Hub) processTransientMsg(client *Client, message *ClientMessage) {
 		}
 
 		if msg.Value == nil {
-			room.SetTransientData(msg.Key, nil)
+			room.SetTransientDataTTL(msg.Key, nil, msg.TTL)
 		} else {
-			room.SetTransientData(msg.Key, *msg.Value)
+			room.SetTransientDataTTL(msg.Key, *msg.Value, msg.TTL)
 		}
 	case "remove":
 		if !isAllowedToUpdateTransientData(session) {
@@ -1985,7 +2066,7 @@ func (h *Hub) isInSameCallRemote(ctx context.Context, senderSession *ClientSessi
 		return false
 	}
 
-	var result int32
+	var result atomic.Bool
 	var wg sync.WaitGroup
 	rpcCtx, cancel := context.WithCancel(ctx)
 	defer cancel()
@@ -2005,12 +2086,12 @@ func (h *Hub) isInSameCallRemote(ctx context.Context, senderSession *ClientSessi
 			}
 
 			cancel()
-			atomic.StoreInt32(&result, 1)
+			result.Store(true)
 		}(client)
 	}
 	wg.Wait()
 
-	return atomic.LoadInt32(&result) != 0
+	return result.Load()
 }
 
 func (h *Hub) isInSameCall(ctx context.Context, senderSession *ClientSession, recipientSessionId string) bool {
@@ -2277,34 +2358,6 @@ func getRealUserIP(r *http.Request) string {
 	return r.RemoteAddr
 }
 
-func (h *Hub) lookupClientCountry(client *Client) string {
-	ip := net.ParseIP(client.RemoteAddr())
-	if ip == nil {
-		return noCountry
-	}
-
-	for overrideNet, country := range h.geoipOverrides {
-		if overrideNet.Contains(ip) {
-			return country
-		}
-	}
-
-	if ip.IsLoopback() {
-		return loopback
-	}
-
-	country, err := h.geoip.LookupCountry(ip)
-	if err != nil {
-		log.Printf("Could not lookup country for %s: %s", ip, err)
-		return unknownCountry
-	}
-
-	if country == "" {
-		return unknownCountry
-	}
-	return country
-}
-
 func (h *Hub) serveWs(w http.ResponseWriter, r *http.Request) {
 	addr := getRealUserIP(r)
 	agent := r.Header.Get("User-Agent")
@@ -2323,23 +2376,47 @@ func (h *Hub) serveWs(w http.ResponseWriter, r *http.Request) {
 
 	h.processNewClient(client)
 	go func(h *Hub) {
-		atomic.AddUint32(&h.writePumpActive, 1)
-		defer atomic.AddUint32(&h.writePumpActive, ^uint32(0))
+		h.writePumpActive.Add(1)
+		defer h.writePumpActive.Add(-1)
 		client.WritePump()
 	}(h)
 	go func(h *Hub) {
-		atomic.AddUint32(&h.readPumpActive, 1)
-		defer atomic.AddUint32(&h.readPumpActive, ^uint32(0))
+		h.readPumpActive.Add(1)
+		defer h.readPumpActive.Add(-1)
 		client.ReadPump()
 	}(h)
 }
 
 func (h *Hub) OnLookupCountry(client *Client) string {
-	if h.geoip == nil {
-		return unknownCountry
+	ip := net.ParseIP(client.RemoteAddr())
+	if ip == nil {
+		return noCountry
 	}
 
-	return h.lookupClientCountry(client)
+	for overrideNet, country := range h.geoipOverrides {
+		if overrideNet.Contains(ip) {
+			return country
+		}
+	}
+
+	if ip.IsLoopback() {
+		return loopback
+	}
+
+	country := unknownCountry
+	if h.geoip != nil {
+		var err error
+		country, err = h.geoip.LookupCountry(ip)
+		if err != nil {
+			log.Printf("Could not lookup country for %s: %s", ip, err)
+			return unknownCountry
+		}
+
+		if country == "" {
+			country = unknownCountry
+		}
+	}
+	return country
 }
 
 func (h *Hub) OnClosed(client *Client) {
