@@ -244,6 +244,15 @@ func (r *Room) processBackendRoomRequestRoom(message *BackendServerRoomRequest)
 		r.publishRoomMessage(message.Message)
 	case "switchto":
 		r.publishSwitchTo(message.SwitchTo)
+	case "transient":
+		switch message.Transient.Action {
+		case TransientActionSet:
+			r.SetTransientDataTTL(message.Transient.Key, message.Transient.Value, message.Transient.TTL)
+		case TransientActionDelete:
+			r.RemoveTransientData(message.Transient.Key)
+		default:
+			log.Printf("Unsupported transient action in room %s: %+v", r.Id(), message.Transient)
+		}
 	default:
 		log.Printf("Unsupported backend room request with type %s in %s: %+v", message.Type, r.Id(), message)
 	}
@@ -1059,6 +1068,10 @@ func (r *Room) SetTransientData(key string, value interface{}) {
 	r.transientData.Set(key, value)
 }
 
+func (r *Room) SetTransientDataTTL(key string, value interface{}, ttl time.Duration) {
+	r.transientData.SetTTL(key, value, ttl)
+}
+
 func (r *Room) RemoveTransientData(key string) {
 	r.transientData.Remove(key)
 }
