@@ -28,7 +28,6 @@ import (
 	"net/url"
 	"sync/atomic"
 	"time"
-	"unsafe"
 )
 
 const (
@@ -38,19 +37,18 @@ const (
 )
 
 type VirtualSession struct {
-	inCall uint32
-
 	hub       *Hub
 	session   *ClientSession
 	privateId string
 	publicId  string
 	data      *SessionIdData
-	room      unsafe.Pointer
+	room      atomic.Pointer[Room]
 
 	sessionId string
 	userId    string
 	userData  *json.RawMessage
-	flags     uint32
+	inCall    Flags
+	flags     Flags
 	options   *AddSessionOptions
 }
 
@@ -69,7 +67,6 @@ func NewVirtualSession(session *ClientSession, privateId string, publicId string
 		sessionId: msg.SessionId,
 		userId:    msg.UserId,
 		userData:  msg.User,
-		flags:     msg.Flags,
 		options:   msg.Options,
 	}
 
@@ -82,6 +79,9 @@ func NewVirtualSession(session *ClientSession, privateId string, publicId string
 	} else if !session.HasFeature(ClientFeatureInternalInCall) {
 		result.SetInCall(FlagInCall | FlagWithPhone)
 	}
+	if msg.Flags != 0 {
+		result.SetFlags(msg.Flags)
+	}
 
 	return result, nil
 }
@@ -99,7 +99,7 @@ func (s *VirtualSession) ClientType() string {
 }
 
 func (s *VirtualSession) GetInCall() int {
-	return int(atomic.LoadUint32(&s.inCall))
+	return int(s.inCall.Get())
 }
 
 func (s *VirtualSession) SetInCall(inCall int) bool {
@@ -107,16 +107,7 @@ func (s *VirtualSession) SetInCall(inCall int) bool {
 		inCall = 0
 	}
 
-	for {
-		old := atomic.LoadUint32(&s.inCall)
-		if old == uint32(inCall) {
-			return false
-		}
-
-		if atomic.CompareAndSwapUint32(&s.inCall, old, uint32(inCall)) {
-			return true
-		}
-	}
+	return s.inCall.Set(uint32(inCall))
 }
 
 func (s *VirtualSession) Data() *SessionIdData {
@@ -144,11 +135,18 @@ func (s *VirtualSession) UserData() *json.RawMessage {
 }
 
 func (s *VirtualSession) SetRoom(room *Room) {
-	atomic.StorePointer(&s.room, unsafe.Pointer(room))
+	s.room.Store(room)
+	if room != nil {
+		if err := s.hub.roomSessions.SetRoomSession(s, s.PublicId()); err != nil {
+			log.Printf("Error adding virtual room session %s: %s", s.PublicId(), err)
+		}
+	} else {
+		s.hub.roomSessions.DeleteRoomSession(s)
+	}
 }
 
 func (s *VirtualSession) GetRoom() *Room {
-	return (*Room)(atomic.LoadPointer(&s.room))
+	return s.room.Load()
 }
 
 func (s *VirtualSession) LeaveRoom(notify bool) *Room {
@@ -205,8 +203,8 @@ func (s *VirtualSession) notifyBackendRemoved(room *Room, session *ClientSession
 
 		if response.Type == "error" {
 			virtualSessionId := GetVirtualSessionId(s.session, s.PublicId())
-			log.Printf("Could not leave virtual session %s at backend %s: %+v", virtualSessionId, s.BackendUrl(), response.Error)
-			if session != nil && message != nil {
+			if session != nil && message != nil && (response.Error == nil || response.Error.Code != "no_such_room") {
+				log.Printf("Could not leave virtual session %s at backend %s: %+v", virtualSessionId, s.BackendUrl(), response.Error)
 				reply := message.NewErrorServerMessage(NewError("remove_failed", response.Error.Error()))
 				session.SendMessage(reply)
 			}
@@ -242,50 +240,19 @@ func (s *VirtualSession) SessionId() string {
 }
 
 func (s *VirtualSession) AddFlags(flags uint32) bool {
-	for {
-		old := atomic.LoadUint32(&s.flags)
-		if old&flags == flags {
-			// Flags already set.
-			return false
-		}
-		newFlags := old | flags
-		if atomic.CompareAndSwapUint32(&s.flags, old, newFlags) {
-			return true
-		}
-		// Another thread updated the flags while we were checking, retry.
-	}
+	return s.flags.Add(flags)
 }
 
 func (s *VirtualSession) RemoveFlags(flags uint32) bool {
-	for {
-		old := atomic.LoadUint32(&s.flags)
-		if old&flags == 0 {
-			// Flags not set.
-			return false
-		}
-		newFlags := old & ^flags
-		if atomic.CompareAndSwapUint32(&s.flags, old, newFlags) {
-			return true
-		}
-		// Another thread updated the flags while we were checking, retry.
-	}
+	return s.flags.Remove(flags)
 }
 
 func (s *VirtualSession) SetFlags(flags uint32) bool {
-	for {
-		old := atomic.LoadUint32(&s.flags)
-		if old == flags {
-			return false
-		}
-
-		if atomic.CompareAndSwapUint32(&s.flags, old, flags) {
-			return true
-		}
-	}
+	return s.flags.Set(flags)
 }
 
 func (s *VirtualSession) Flags() uint32 {
-	return atomic.LoadUint32(&s.flags)
+	return s.flags.Get()
 }
 
 func (s *VirtualSession) Options() *AddSessionOptions {
@@ -308,6 +275,41 @@ func (s *VirtualSession) ProcessAsyncSessionMessage(message *AsyncMessage) {
 				}
 				s.session.ProcessAsyncSessionMessage(message)
 			}
+		case "event":
+			if room := s.GetRoom(); room != nil &&
+				message.Message.Event.Target == "roomlist" &&
+				message.Message.Event.Type == "disinvite" &&
+				message.Message.Event.Disinvite != nil &&
+				message.Message.Event.Disinvite.RoomId == room.Id() {
+				log.Printf("Virtual session %s was disinvited from room %s, hanging up", s.PublicId(), room.Id())
+				payload := map[string]interface{}{
+					"type": "hangup",
+					"hangup": map[string]string{
+						"reason": "disinvited",
+					},
+				}
+				data, err := json.Marshal(payload)
+				if err != nil {
+					log.Printf("could not marshal control payload %+v: %s", payload, err)
+					return
+				}
+
+				s.session.ProcessAsyncSessionMessage(&AsyncMessage{
+					Type:     "message",
+					SendTime: message.SendTime,
+					Message: &ServerMessage{
+						Type: "control",
+						Control: &ControlServerMessage{
+							Recipient: &MessageClientMessageRecipient{
+								Type:      "session",
+								SessionId: s.SessionId(),
+								UserId:    s.UserId(),
+							},
+							Data: (*json.RawMessage)(&data),
+						},
+					},
+				})
+			}
 		case "control":
 			if message.Message.Control != nil &&
 				message.Message.Control.Recipient != nil &&
